"""

Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import random

import cereja.mathtools
from cereja.array import array_gen, Matrix, array_randn


class HyperParam(object):
    def __init__(self):
        self.n_steps = 1000
        self.episode_length = 1000
        self.learning_rate = 0.02
        self.n_directions = 16
        self.n_best_directions = 16
        assert self.n_best_directions <= self.n_directions, "n_best_directions > n_directions"
        self.noise = 0.03
        self.seed = 1
        self.env_name = 'HalfCheetahBulletEnv-v0'


hp = HyperParam()


class Normalizer(object):
    def __init__(self, n_inputs):
        self.n = Matrix(array_gen((n_inputs,), 0.0))
        self.mean = Matrix(array_gen((n_inputs,), 0.0))
        self.mean_diff = Matrix(array_gen((n_inputs,), 0.0))
        self.var = Matrix(array_gen((n_inputs,), 0.0))

    def observe(self, state):
        self.n += 1.
        last_mean = self.mean.copy()
        self.mean += (state - self.mean) / self.n
        self.mean_diff += (state - last_mean) * (state - self.mean)
        self.var = (self.mean_diff / self.n)  # 1e-2 need clip

    def normalize(self, inputs):
        observe_mean = self.mean
        obs_std = self.var.sqrt()
        return (inputs - observe_mean) / obs_std


class Policy(object):
    def __init__(self, input_size: int, output_size: int):
        self.weights = Matrix(array_gen((output_size, input_size), 0.0))

    def evaluate(self, input_, delta=None, direction=None):
        if direction is None:
            return self.weights.dot(input_)
        elif direction == 'posittive':
            return cereja.mathtools.dot(input_)
        else:
            return cereja.mathtools.dot(input_)

    def sample_deltas(self):
        return [Matrix(array_randn(self.weights.shape)) for _ in range(hp.n_directions)]

    def update(self, rollouts, sigma_r):
        step = Matrix(array_gen(self.weights.shape, 0.0))
        for r_pos, r_neg, d in rollouts:
            step += (r_pos - r_neg) * d
        self.weights += hp.learning_rate / (hp.n_best_directions * sigma_r) * step


def explore(env, normalizer, policy, direction=None, delta=None):
    state = env.reset()
    done = False
    num_plays = 0.
    sum_rewards = 0
    while not done and num_plays < hp.episode_length:
        normalizer.observe(state)
        state = normalizer.normalize(state)
        action = policy.evaluate(state, delta, direction)
        state, reward, done, _ = env.step(action)
        reward = max(min(reward, 1), -1)
        sum_rewards += reward
        num_plays += 1
    return sum_rewards


def train(env, policy: Policy, normalizer, hp: HyperParam):
    for step in range(hp.n_steps):
        deltas = policy.sample_deltas()
        positive_rewards = [0] * hp.n_directions
        negative_rewards = [0] * hp.n_directions
        for k in range(hp.n_directions):
            positive_rewards[k] = explore(env, normalizer, policy, direction='positive', delta=deltas[k])

        for k in range(hp.n_directions):
            negative_rewards[k] = explore(env, normalizer, policy, direction='negattive', delta=deltas[k])

        all_rewards = Matrix((positive_rewards + negative_rewards))
        sigma_r = all_rewards.std()
        score = {k: max(r_pos, r_neg) for k, (r_pos, r_neg) in enumerate(zip(positive_rewards, negative_rewards))}
        order = sorted(score.keys(), key=lambda x: score[x], reverse=True)[:hp.n_best_directions]
        rollouts = [(positive_rewards[k], negative_rewards[k], deltas[k]) for k in order]
        policy.update(rollouts, sigma_r)
        reward_evaluation = explore(env, normalizer, policy)
        print(f'Step: {step} - Reward: {reward_evaluation}')


random.seed(hp.seed)



"""
Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import math
from typing import Tuple

from cereja.config.cj_types import Number

__all__ = ['imc', 'proportional', 'estimate', 'percent', 'theta_angle', 'distance_between_points']


def imc(weight: float, height: float) -> Tuple[float, str]:
    _imc = weight / (height ** 2)
    _imc = _imc if _imc > 0.1 else _imc * 10000  # convert if send in meters.
    if _imc < 18.5:
        grade = 'Underweight'
    elif 18.5 <= _imc <= 24.9:
        grade = 'Normal weight'
    elif 25 <= _imc <= 29.9:
        grade = 'Overweight'
    elif 30 <= _imc <= 34.9:
        grade = 'Obesity grade 1'
    elif 35 <= _imc <= 39.9:
        grade = 'Obesity grade 2'
    else:
        grade = 'Obesity grade 3'
    return _imc, grade


def proportional(value, pro_of_val: int):
    return (pro_of_val / 100) * value


def estimate(from_: Number, to: Number, based: Number, ndigits=2) -> Number:
    if from_ > 0:
        based = based or 1
        return round((based / from_) * to - based, ndigits)
    return float('NaN')


def percent(from_: Number, to: Number, ndigits=2) -> Number:
    to = to or 1
    return round((from_ / to) * 100, ndigits)


def theta_angle(u: Tuple[float, float], v: Tuple[float, float]) -> float:
    """
    Calculates and returns theta angle between two vectors

    e.g usage:
    >>> u = (2,2)
    >>> v = (0, -2)
    >>> theta_angle(u, v)
    135.0
    """
    x1, y1 = u
    x2, y2 = v
    return math.degrees(math.acos((x1 * x2 + y1 * y2) / (math.sqrt(x1 ** 2 + y1 ** 2) * math.sqrt(x2 ** 2 + y2 ** 2))))


def distance_between_points(u: Tuple[float, float], v: Tuple[float, float]):
    x1, y1 = u
    x2, y2 = v
    return math.sqrt(((x2 - x1) ** 2) + ((y2 - y1) ** 2))



import unittest
from collections import OrderedDict

from cereja import utils, rescale_values
from cereja.utils.decorators import singleton


class UtilsTest(unittest.TestCase):

    def test_camel_to_snake(self):
        pass

    def test_class_methods(self):
        pass

    def test_combine_with_all(self):
        pass

    def test_fill(self):
        pass

    def test_get_attr_if_exists(self):
        pass

    def test_get_implements(self):
        pass

    def test_get_instances_of(self):
        pass

    def test_import_string(self):
        pass

    def test_install_if_not(self):
        pass

    def test_invert_dict(self):
        data = {"s": 0, "y": 1, "v": 2, "l": 3, "i": 4, "p": 5, "b": 6, "z": 7, "c": 8, "a": 9, "k": 10, "e": 11,
                "d": 12, "j": 13, "x": 14, "u": 15, "o"
                :    16, "n": 17, "t": 18, "f": 19, "g": 20, "h": 21, "r": 22, "w": 23, "m": 24, "q": 25}
        expected = {0:  's', 1: 'y', 2: 'v', 3: 'l', 4: 'i', 5: 'p', 6: 'b', 7: 'z', 8: 'c', 9: 'a', 10: 'k', 11: 'e',
                    12: 'd', 13: 'j', 14: 'x', 15: 'u', 16: 'o', 17: 'n', 18: 't', 19: 'f', 20: 'g', 21: 'h', 22: 'r',
                    23: 'w', 24: 'm', 25: 'q'}
        self.assertDictEqual(utils.invert_dict(data), expected)

        data = {1: 40, 2: 30, 3: 40, 30: 2}
        expected = {40: [1, 3], 30: 2, 2: 30}
        self.assertEqual(utils.invert_dict(data), expected)

    def test_logger_level(self):
        pass

    def test_module_references(self):
        pass

    def test_obj_repr(self):
        class Repr:
            def __init__(self, a, b, c):
                self.a = a
                self.b = b
                self.c = c

        expected = "Repr (\n    a = 1,\n    b = dict(oi:<class 'int'> ...),\n    c = oi\n    )"
        self.assertEqual(utils.obj_repr(Repr(1, {'oi': 1}, 'oi')), expected)

    def test_rescale_values(self):
        value = list(range(10))
        expected = [0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9]
        self.assertEqual(rescale_values(value, 21), expected)
        expected = [0, 0, 0, 1, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9]
        self.assertEqual(rescale_values(value, 22), expected)
        expected = [0, 0, 'joab', 1, 1, 'joab', 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9]
        self.assertEqual(rescale_values(value, 22, fill_with='joab'), expected)

    def test_can_do(self):
        class _A:
            def _m1(self):
                pass

            def __m2(self):
                pass

            def m(self):
                pass

            @property
            def m1(self):
                return self._m1()

        a_obj = _A()

        self.assertEqual(utils.can_do(a_obj), ['m', 'm1'])

    def test_sample(self):
        data = ['The', 'Cereja', 'is', 'for', 'everyone']

        self.assertEqual(utils.sample(data, 2), ['The', 'Cereja'])
        self.assertEqual(utils.sample(data, 3), ['The', 'Cereja', 'is'])
        self.assertEqual(utils.sample('Cereja'), list('Cereja'))  # test sent string
        self.assertEqual(utils.sample(b'Cereja'), list(b'Cereja'))  # test sent bytes
        self.assertEqual(utils.sample([12], 2), [12])  # test sent number

        data = {'The': 'Cereja', 'is': 'for', 'everyone': None}
        self.assertEqual(utils.sample(data, 2), {'The': 'Cereja', 'is': 'for'})

    def test_set_log_level(self):
        pass

    def test_string_to_literal(self):
        pass

    def test_time_format(self):
        pass

    def test_truncate(self):
        self.assertEqual(utils.truncate("Cereja is fun.", k=3), 'Cer...')
        self.assertEqual(utils.truncate(b"Cereja is fun.", k=3), b'Cer...')
        self.assertEqual(utils.truncate("Cereja is fun.", k=-1), "Cereja is fun.")
        self.assertEqual(utils.truncate("Cereja is fun.", k=1000), "Cereja is fun.")

        self.assertRaises(AssertionError, utils.truncate, 1123)

    def test_type_table_of(self):
        pass

    def test_sort_dict(self):
        val = {0: 1, 1: 2, 2: 1, 3: 4, 4: 1, 5: 43, 6: 1, 7: 10, 8: 22, 9: 0}
        self.assertDictEqual(utils.sort_dict(val), OrderedDict(
                [(0, 1), (1, 2), (2, 1), (3, 4), (4, 1), (5, 43), (6, 1), (7, 10), (8, 22), (9, 0)]))
        self.assertDictEqual(utils.sort_dict(val, by_values=True), OrderedDict(
                [(9, 0), (0, 1), (2, 1), (4, 1), (6, 1), (1, 2), (3, 4), (7, 10), (8, 22), (5, 43)]))
        self.assertDictEqual(utils.sort_dict(val, by_values=True, reverse=True), OrderedDict(
                [(5, 43), (8, 22), (7, 10), (3, 4), (1, 2), (0, 1), (2, 1), (4, 1), (6, 1), (9, 0)]))

        self.assertDictEqual(utils.sort_dict(val, by_keys=True), OrderedDict(
                [(0, 1), (1, 2), (2, 1), (3, 4), (4, 1), (5, 43), (6, 1), (7, 10), (8, 22), (9, 0)]))
        self.assertDictEqual(utils.sort_dict(val, by_keys=True, reverse=True), OrderedDict(
                [(9, 0), (8, 22), (7, 10), (6, 1), (5, 43), (4, 1), (3, 4), (2, 1), (1, 2), (0, 1)]))

    def test_dict_append(self):
        my_dict = {}
        utils.dict_append(my_dict, 'key_eg', 1, 2, 3, 4, 5, 6)
        self.assertDictEqual(my_dict, {'key_eg': [1, 2, 3, 4, 5, 6]})
        my_dict = utils.dict_append(my_dict, 'key_eg', [1, 2])

        self.assertDictEqual(my_dict, {'key_eg': [1, 2, 3, 4, 5, 6, [1, 2]]})

    def test_to_tuple(self):
        data = [[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]]
        expected = (
            ((0.0, 0.0, 0.0), (0.0, 0.0, 0.0), (0.0, 0.0, 0.0)), ((0.0, 0.0, 0.0), (0.0, 0.0, 0.0), (0.0, 0.0, 0.0)),
            ((0.0, 0.0, 0.0), (0.0, 0.0, 0.0), (0.0, 0.0, 0.0)))

        self.assertEqual(utils.to_tuple(data), expected)

        data = {1: [1, 2, [3, 100, [1000, 30, [12, 3]]], {1: 2, 3: 4}, 4], 2: [1, 2, 3, 4], 3: {1, 2, 3}}
        expected = ((1, (1, 2, (3, 100, (1000, 30, (12, 3))), ((1, 2), (3, 4)), 4)), (2, (1, 2, 3, 4)), (3, (1, 2, 3)))

        self.assertEqual(utils.to_tuple(data), expected)


class CjTestTest(unittest.TestCase):

    def test_add_check(self):
        pass

    def test_build_test(self):
        pass

    def test_check_all(self):
        pass

    def test_check_attr(self):
        pass

    def test_parse_attr(self):
        pass

    def test_remove_check(self):
        pass


class SourceTest(unittest.TestCase):

    def test_save(self):
        pass


class DecoratorTest(unittest.TestCase):

    def test_depreciation(self):
        pass

    def test_singleton(self):
        @singleton
        class A:
            def __init__(self, a, b, c=None):
                self.a = a
                self.b = b
                self.c = c

        obj_expected = A(1, b=2, c=3)
        obj_new = A(1, 2)
        self.assertEqual(obj_expected, obj_new, msg="Error on decorator singleton.")

    def test_synchronized(self):
        pass

    def test_thread_safe_generator(self):
        pass

    def test_time_exec(self):
        pass


if __name__ == '__main__':
    unittest.main()



import unittest
from cereja import hashtools


class HashtoolsTest(unittest.TestCase):

    def test_base64_decode(self):
        self.assertEqual(hashtools.base64_decode(b'YWJjZGVmIGcgaCBpIGo=', eval_str=True), 'abcdef g h i j')
        self.assertEqual(hashtools.base64_decode(b'eydvaSc6ICd0dWRvIGJlbSd9', eval_str=True), {'oi': 'tudo bem'})
        self.assertEqual(hashtools.base64_decode(b'YWJjZGVmIGcgaCBpIGo='), b'abcdef g h i j')

    def test_base64_encode(self):
        self.assertEqual(hashtools.base64_encode('abcdef g h i j'), b'YWJjZGVmIGcgaCBpIGo=')
        self.assertEqual(hashtools.base64_encode({'oi': 'tudo bem'}), b'eydvaSc6ICd0dWRvIGJlbSd9')

    def test_is_base64(self):
        pass

    def test_md5(self):
        pass


if __name__ == '__main__':
    unittest.main()
